/***********************************************************

    hbs reduction subroutine: hbsiqupt.c

                      1997/09/10  K. Kawabata
                      1998/10/15  K. Kawabata  weighted-bin
                      1998/11/23  K. Kawabata  mean-error
                      2005/09/08  M. Isogai    Andor CCD

    I, q, u spectrum   --->  TABLE of I, q, u, p and theta 

    usage: % hbsiqupt inlist binfile photfile weight-file outfn

    binfile -- table of binning parameters
   photfile -- photometric parameter file (made by hbsphfit3.cl)
 weightfile -- weight (unbinned count file: *1.xy)
      outfn -- file name of output table

   Format of 'inlist' file:

      fn_kn
      infn1 infn2 flag1
      infn3 infn4 flag2
      infn5 infn6 flag3
      mode
      upfile
      gtfile
      spfile

      fn_kn
      infn1 -- I spectrum (FITS format)
      infn2 -- Spectrum of error of I (FITS format)
      flag1 -- flag of existence of I (0:exist -1:not exist)
      infn3 -- q spectrum (FITS format)
      infn4 -- Spectrum of error of q (FITS format)
      flag2 -- flag of existence of q (0:exist -1:not exist)
      infn5 -- u spectrum (FITS format)
      infn6 -- Spectrum of error of u (FITS format)
      flag3 -- flag of existence of u (0:exist -1:not exist)
       mode -- 0:celetial(using gtfile, spfile) 1:instrumental
     upfile -- UP correction table (made by hbsupave)
     gtfile -- GT correction table (made by hbsgtave)
     spfile -- Position angle defference file (made by hbsspave)


    Output table contains
       column     contents
        1     wavelength (angstrom)
        2     I (count)
        3     q (=Q/I) (%)
        4     error of q
        5     u (=U/I) (%)
        6     error of q
        7     P (%)
        8     error of P
        9     position angle (degree)
       10     error of position angle

       if( flag2==-1 && flag3==-1 )

************************************************************/

#include<stdio.h>
#include<string.h>
#include<math.h>
#include<stdlib.h>
#include<unistd.h>
#include"fitsio.h"
#include"hbsinterpol.c"

#include"hbsdefine.h"

char detect[50];

int scanfile( char *infile, float count[]  ){

  fitsfile *fp;
  int i, anynul, status;
  char buf[100];
  float nulval = 0.;
  long naxis1, naxis2, npix;
  char comment[100];

  status = 0;

  if( fits_open_file( &fp, infile, READONLY, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS1", &naxis1,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS2", &naxis2,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( naxis1 != NX3 || naxis2 != 1 ){
    printf( "error: definition of NX3 or NY(=1) in 'hbsdefine.h'\n" );
    return( -1 );
  }
  npix = naxis1 * naxis2;

  if( fits_read_img( fp, TFLOAT, 1L, npix, &nulval,\
      count, &anynul, &status ) != 0 ){
    return( status );
  }
  if( fits_close_file( fp, &status ) != 0 ){
    return( status );
  }

  return( 0 );

}

int readheads( char *infile, char *keyword  ){

  fitsfile *fpf;
  int status;
  char comment[100];

  status = 0;

  if( fits_open_file( &fpf, infile, READONLY, &status ) != 0 ){
    return( -1 );
  }
  if( fits_read_keyword( fpf, keyword, detect,\
      comment, &status ) != 0 ){
    return( -1 );
  }
  if( fits_close_file( fpf, &status ) != 0 ){
    return( status );
  }
  return( 0 );
}

int readheadi( char *infile, char *keyword  ){

  fitsfile *fpf;
  int status;
  int value=0;
  char comment[100];

  status = 0;

  if( fits_open_file( &fpf, infile, READONLY, &status ) != 0 ){
    return( -1 );
  }
  if( fits_read_key( fpf, TINT, keyword, &value,\
      comment, &status ) != 0 ){
    return( -1 );
  }
  if( fits_close_file( fpf, &status ) != 0 ){
    return( status );
  }
  return( value );
}

double readheadd( char *infile, char *keyword  ){

  fitsfile *fpf;
  int status;
  double value;
  char comment[100];

  status = 0;

  if( fits_open_file( &fpf, infile, READONLY, &status ) != 0 ){
    return( -1 );
  }
  if( fits_read_key( fpf, TDOUBLE, keyword, &value,\
      comment, &status ) != 0 ){
    return( -1 );
  }
  if( fits_close_file( fpf, &status ) != 0 ){
    return( status );
  }
  return( value );
}

int main( int argc, char *argv[]){

  int thetamin, thetamax;
  double crval, cdelt, crpix;
  float I[NX3], Isig[NX3], Q[NX3], Qsig[NX3], U[NX3], Usig[NX3];
  double i, isig, q, q_tmp, qsig, qsig_tmp, u, u_tmp, usig, usig_tmp;
  double p, psig, theta, thetasig, lambda;
  float n_qu[NX3];
  double ONEDEG = 3.14159265358979 / 180.0;
  double dtheta[NX3], gtp[NX3], lgt[NX3];
  double lam_c[NX3], lam_eff[NX3], dlam_eff[NX3];
  char lam_w[NX3][15];
  double lup[NX3], upq[NX3], upu[NX3];
  double a, rottheta; /* position angle rotation from spfile */
  double xweit[NX3], yweit[NX3]; /* weight for average (e.g. real count) */
  int nweit;
  int pix;
  double xphot[NX3], yphot[NX3];
  int pregain, postgain;
  double exptime, tmp;
  int nphot;

  FILE *fp;
  char buf[200], last_buf[200];
  char list1[50]="", list2[50]="", list3[20]="", list4[20]="";
  char fn_iave[50], fn_qave[50], fn_uave[50];
  char fn_isig[50], fn_qsig[50], fn_usig[50];
  char fn_kn[50];
  char upfile[50], gtfile[50], spfile[50];
  char binfile[50], weitfn[50], refdir[50], photfile[50];
  char detector[50];
  int flag_mode;
  int flag_i, flag_q, flag_u;
  char outfile[50];
  int npix;

/***** confirmation of input file *****/

  if( argc != 7  ){
    printf( "usage: hbsiqupt Iqu-list binfile phot-fn weight-fn refdir outfn\n" );
    for( pix=0; pix<=(argc-1); pix++ ){
      printf( "%2d  %s\n", pix, argv[pix] );
    }
    exit( -1 );
  }

  strcpy( buf, argv[1] );
  strcpy( binfile, argv[2] );
  strcpy( photfile, argv[3] );
  strcpy( weitfn, argv[4] );
  strcpy( refdir, argv[5] );
  strcpy( outfile, argv[6] );

/*** read I, q, u images and its flags */

  if( ( fp = fopen( buf, "rt" ) ) == NULL ){
    printf( "cannot open Iqu-list %s", buf );
    exit( -1 );
  }
  fgets( buf, 200, fp );
  sscanf( buf, "%s", fn_kn );
  fgets( buf, 200, fp );
  sscanf( buf, "%s %s %d", fn_iave, fn_isig, &flag_i );
  fgets( buf, 200, fp );
  sscanf( buf, "%s %s %d", fn_qave, fn_qsig, &flag_q );
  fgets( buf, 200, fp );
  sscanf( buf, "%s %s %d", fn_uave, fn_usig, &flag_u );
  fgets( buf, 200, fp );
  sscanf( buf, "%d", &flag_mode );
  fgets( buf, 200, fp );
  sscanf( buf, "%s", upfile );
  fgets( buf, 200, fp );
  sscanf( buf, "%s", gtfile );
  fgets( buf, 200, fp );
  sscanf( buf, "%s", spfile );

  fclose( fp );


/* read photometric parameter file */

  if( strcmp( photfile, "NONE" ) != 0 ){
    pregain = -1;
    postgain = -1;
    exptime = -1.0;

    tmp = readheads( fn_iave, "DETECTOR" );
    strcpy(detector,detect);

    if (strcmp(detector,"'1024x1024 TI CCD'") == 0 || strcmp(detector,"'T5 CRYOCAM (TC215)'") == 0 || strcmp(detector,"'S5 CRYOCAM (SITe)'") == 0)
      {
	pregain = readheadi( fn_iave, "PREGAIN" );
	if( pregain == -1 ){
	  printf( "Cannot find paramater 'PREGAIN' in %s\n", fn_iave );
	  exit( -1 );
	}
	postgain = readheadi( fn_iave, "POSTGAIN" );
	if( postgain == -1 ){
	  printf( "Cannot find paramater 'POSTGAIN' in %s\n", fn_iave );
	  exit( -1 );
	}
      }
    else if (strcmp(detector,"'Andor DV440'") == 0 )
      {
	pregain=1;
	postgain=1;
      }
    else
      {
	printf("Warning! header:detector is %s\n",detector);
      }

    exptime = readheadi( fn_iave, "EXPTIME" );
    if( exptime == -1.0 ){
      printf( "Cannot find paramater 'EXPTIME' in %s\n", fn_iave );
      exit( -1 );
    }

    nphot = -1;

    if( ( fp = fopen( photfile, "rt" ) ) == NULL ){
      printf( "Photmetric parameter file '%s' not found.\n", photfile );
      exit( -1 );
    }
    while( feof( fp ) == 0 ){
      fgets( buf, 200, fp );
      if( strcmp( buf, last_buf ) == 0 )
	continue;
      nphot++;
      if( nphot >= NX3 ){
	printf( "n_phot exceeds NX3.\n" );
	break;
      }
      sscanf( buf, "%lf %lf %lf %lf %lf %lf", &xphot[nphot], &a, &a,\
	     &a, &yphot[nphot], &a );
      yphot[nphot] *= ( 2.9979e+18 / ( xphot[nphot] * xphot[nphot]\
				      * GEADU * pregain * postgain ) );
      strcpy( last_buf, buf );
    }
    fclose(fp);
  }
  else{
    nphot = 1;
    xphot[0] = 0.000;
    xphot[1] = 20000.000;
    yphot[0] = 1.000;
    yphot[1] = 1.000;
  }

/*** read weight factor */

  if( strcmp( weitfn, "NONE" ) != 0 ){
    nweit = -1;
    if( ( fp = fopen( weitfn, "rt" ) ) == NULL ){
      printf( "cannot open weight-file %s", weitfn );
      exit( -1 );
    }
    while( feof( fp ) == 0 ){
      fgets( buf, 200, fp );
      if( strcmp( last_buf, buf ) == 0 )
	continue;
      nweit++;
      sscanf( buf, "%lf %lf", &xweit[nweit], &yweit[nweit] );
      strcpy( last_buf, buf );
    }
    fclose( fp );
    xweit[0] = 0.0;
    xweit[NX3-1] = 20000.0; /* for the edge pixel */
  }
  else{
    nweit = 2;
    xweit[0] = 0.000;
    xweit[1] = 20000.000;
    yweit[0] = 1.000;
    yweit[1] = 1.000;
  }

  /*** read n_qu, I, q, u and their errors ***/

  if( scanfile( fn_kn, n_qu ) != 0 ){
    printf( "Read error in '%s'\n", fn_kn );
    exit( 1 );
  }

  if( flag_i == 0 ){
    if( scanfile( fn_iave, I ) != 0 ){
      printf( "Read error in '%s'\n", fn_iave );
      exit( 1 );
    }
    if( scanfile( fn_isig, Isig ) != 0 ){
      printf( "Read error in '%s'\n", fn_isig );
      exit( 1 );
    }
  }
  else{
    for( pix=0; pix<=(NX3-1); pix+=1 ){
      I[pix] = 0.0;
      Isig[pix] = 0.0;
    }
  }

  if( flag_q == 0 ){
    if( scanfile( fn_qave, Q ) != 0 ){
      printf( "Read error in '%s'\n", fn_qave );
      exit( 1 );
    }
    if( scanfile( fn_qsig, Qsig ) != 0 ){
      printf( "Read error in '%s'\n", fn_qsig );
      exit( 1 );
    }
  }
  else{
    for( pix=0; pix<=(NX3-1); pix+=1 ){
      Q[pix] = 0.0;
      Qsig[pix] = 0.0;
    }
  }

  if( flag_u == 0 ){
    if( scanfile( fn_uave, U ) != 0 ){
      printf( "Read error in '%s'\n", fn_uave );
      exit( 1 );
    }
    if( scanfile( fn_usig, Usig ) != 0 ){
      printf( "Read error in '%s'\n", fn_usig );
      exit( 1 );
    }
  }
  else{
    for( pix=0; pix<=(NX3-1); pix+=1 ){
      U[pix] = 0.0;
      Usig[pix] = 0.0;
    }
  }
  
  if( flag_q == 0 || flag_u == 0 ){

/*** read binning parameters ***/
  
    npix = -1;
    strcpy( last_buf, "" );
    if( ( fp = fopen( binfile, "rt" ) ) == NULL ){
      printf( "cannot open bin-file %s", binfile );
      exit( 1 );
    }
    while( feof( fp ) == 0 ){
      fgets( buf, 200, fp );
      if( strcmp( last_buf, buf ) == 0 )
	continue;
      npix++;
      sscanf( buf, "%lf %s %lf %lf", &lam_c[npix], lam_w[npix],\
	      &lam_eff[npix], &dlam_eff[npix] );
      strcpy( last_buf, buf );
    }
    fclose( fp );
  }
  else{

/*** read crval and cdelt ***/
    npix =(NX3-1);

    crval = SW;
    cdelt = DW;
    crpix = 1;
    /*
    crval = readheadd( fn_iave, "CRVAL1" );
    cdelt = readheadd( fn_iave, "CD1_1" );
    crpix = readheadd( fn_iave, "CRPIX1" );
    */
  }
  
  if( flag_mode == 0 ){

/*** read origin of position angle ***/

    if( ( fp = fopen( spfile, "rt" ) ) == NULL ){
      printf( "cannot open %s", spfile );
      exit( 1 );
    }
    fgets( buf, 200, fp );
    sscanf( buf, "%lf", &rottheta );
    fclose( fp );

/*** read UP correction table ***/
    
    if( ( fp = fopen( upfile, "rt" ) ) == NULL ){
      printf( "cannot open %s", upfile );
      exit( 1 );
    }
    for( pix=0; pix<=(NX3-1); pix++ ){
      if( feof( fp ) != 0 ){
        printf( "File read error in '%s'\n", upfile );
        exit( 1 );
      }
      fgets( buf, 200, fp );
      sscanf( buf, "%lf %lf %lf %lf %lf", &lup[pix],\
	     &upq[pix], &a, &upu[pix], &a );
    }
    fclose( fp );
    if( abs( lup[1] - lup[0] - DW ) > 1.0e6 ){
      printf( "Wavelength width error in '%s'.\n", upfile );
      exit( -1 );
    }

/*** read GT correction table ***/

    if( ( fp = fopen( gtfile, "rt" ) ) == NULL ){
      printf( "cannot open %s", gtfile );
      exit( 1 );
    }
    for( pix=0; pix<=(NX3-1); pix++ ){
      if( feof( fp ) != 0 ){
	printf( "File read error in '%s'\n", gtfile );
	exit( 1 );
      }
      fgets( buf, 200, fp );
      sscanf( buf, "%lf %lf %lf %lf %lf %lf", &lgt[pix],\
	     &gtp[pix], &a, &a, &a, &dtheta[pix] );
      gtp[pix] = ( gtp[pix] / 100.0 );
    }
    fclose( fp );
    if( abs( lgt[1] - lgt[0] - DW ) > 1.0e6 ){
      printf( "Wavelength width error in '%s'.\n", gtfile );
      exit( -1 );
    }
  }    
  else{
    for( pix=0; pix<=(NX3-1); pix++ ){
      upq[pix] = 0.0;
      upu[pix] = 0.0;
      lup[pix] = 0.0;
    }

    rottheta = 0.0;
    for( pix=0; pix<=(NX3-1); pix++ ){
      dtheta[pix] = 0.0;
      gtp[pix] = 0.0;
      lgt[pix] = 0.0;
    }
  }

/*** open output file ***/

  if( ( fp = fopen( outfile, "wt" ) ) == NULL ){
    printf( "cannot open %s", outfile );
    exit( 1 );
  }

/*** theta of middle wavelength (to fix thetamax, thetamin)****/

  if( flag_q == 0 || flag_u == 0 ){
      
    thetamin = 0.0;
    thetamax = 180.0;
      
    pix = npix / 2.0;
    lambda = lam_c[pix];

    q = 100.0 * Q[pix];
    u = 100.0 * U[pix];

    if( flag_mode == 0 ){
      a = hbsaverage( (NX3-1), lgt, gtp, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot );
      q = q / a - hbsaverage( (NX3-1), lup, upq, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot );
      u = u / a - hbsaverage( (NX3-1), lup, upu, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot );

/*
   printf( "UP %7.2fA %+8.3f %+8.3f\n", lambda, hbsaverage( (NX3-1), lup, upq, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot ), hbsaverage( (NX3-1), lup, upu, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot ) );
*/

    }

    p = sqrt( q * q + u * u );

    if( fabs( q ) < 0.000005 ){
      if( fabs( u ) < 0.000005 )
	theta = 0.0;
      else{
	if( u > 0 )  theta = 90.0;
	else      theta = 270.0;
      }
    }
    else{
      if( fabs( u ) < 0.000005 ){
	if( q < 0 )  theta = 180.0;
	else      theta = 0.0;
      }
      else{
	theta = atan( u / q ) / ONEDEG;
	if(q < 0 && u > 0)  theta += 180.0;
	if(q < 0 && u < 0)  theta += 180.0;
	if(q > 0 && u < 0)  theta += 360.0;
      }
    }
    theta /= 2.0;

    if( flag_mode == 0 )
      theta -= ( rottheta + hbsaverage( (NX3-1), lgt, dtheta, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot ) );

    while( theta > thetamax )
      theta -= 180.0;
    while( theta <= thetamin )
      theta += 180.0;
    
    thetamax = theta + 90.0;
    thetamin = theta - 90.0;
  }

  /***** main routine *****/

  for( pix=0; pix<=npix; pix++ ){
      
    i = I[pix];
    isig = Isig[pix];

    if( flag_q == 0 || flag_u == 0 ){
      lambda = lam_c[pix];
/*      lam_w[pix] = lam_w[pix] / DW; */

      q = 100.0 * Q[pix];
      qsig = 100.0 * Qsig[pix];
      u = 100.0 * U[pix];
      usig = 100.0 * Usig[pix];
      
      if( flag_mode == 0 ){

	a = hbsaverage( (NX3-1), lgt, gtp, lambda, lam_w[pix],\
                        refdir, xweit, yweit, nweit, xphot, yphot, nphot );
/*
	if( strcmp( lam_w[pix], "JV" ) == 0 ){
	  printf( "GTscale  %s %+6.4f\n", lam_w[pix], a );
	}
*/

	if( sqrt( q * q + u * u ) < 70.0 ){
	  q = q / a - hbsaverage( (NX3-1), lup, upq, lambda, lam_w[pix],\
	      refdir, xweit, yweit, nweit, xphot, yphot, nphot );
	  u = u / a - hbsaverage( (NX3-1), lup, upu, lambda, lam_w[pix],\
	      refdir, xweit, yweit, nweit, xphot, yphot, nphot );
        }
	else{
	  q = q / a;
	  u = u / a;
	}
/*
	if( strcmp( lam_w[pix], "JV" ) == 0 ){
	  printf( "UPcor  %s  q=%+7.3f u=%+7.3f\n", lam_w[pix],\
		 hbsaverage( (NX3-1), lup, upq, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot ),\
		 hbsaverage( (NX3-1), lup, upu, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot ) );
	}
*/
	qsig = qsig / a;
	usig = usig / a;
/*	qsig = qsig / ( sqrt( n_q - 1 ) * a );
	usig = usig / ( sqrt( n_u - 1 ) * a );
*/
      }

      if( flag_mode == 0 ){

	q_tmp = q;
	u_tmp = u;
	qsig_tmp = qsig;
	usig_tmp = usig;

	a = rottheta + hbsaverage( (NX3-1), lgt, dtheta, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot );

/*
	if( strcmp( lam_w[pix], "JV" ) == 0 ){
	  printf( "GTtheta  %s %+7.3f\n", lam_w[pix],\
		 hbsaverage( (NX3-1), lgt, dtheta, lambda, lam_w[pix], refdir, xweit, yweit, nweit, xphot, yphot, nphot ) );
	}
*/
        q = q_tmp * cos( 2.0 * a * ONEDEG ) + u_tmp * sin( 2.0 * a * ONEDEG );
        u = -q_tmp * sin( 2.0 * a * ONEDEG ) + u_tmp * cos( 2.0 * a * ONEDEG );
        qsig = sqrt( pow( ( qsig_tmp * cos( 2.0 * a * ONEDEG ) ), 2.0 )
		    + pow( ( usig_tmp * sin( 2.0 * a * ONEDEG ) ), 2.0 ) );
        usig = sqrt( pow( ( qsig_tmp * sin( 2.0 * a * ONEDEG ) ), 2.0 )
		    + pow( ( usig_tmp * cos( 2.0 * a * ONEDEG ) ), 2.0 ) );
/*        qsig = usig_tmp;
        usig = qsig_tmp;*/
      }


      psig = -1.0;
      thetasig = -1.0;
      p = sqrt( q * q + u * u );

      if( fabs( q ) < 0.000005 ){
	if( fabs( u ) < 0.000005 ){
	  theta = 0.0;
	  psig = 0.0;
	  thetasig = 0.0;
	}
	else{
	  if( u > 0 )  theta = 90.0;
	  else      theta = 270.0;
	}
      }
      else{
	if( fabs( u ) < 0.000005 ){
	  if( q < 0 )  theta = 180.0;
	  else      theta = 0.0;
	}
	else{
	  theta = atan( u / q ) / ONEDEG;
	  if(q < 0 && u > 0)  theta += 180.0;
	  if(q < 0 && u < 0)  theta += 180.0;
	  if(q > 0 && u < 0)  theta += 360.0;
	}
      }
      theta /= 2.0;

/*      q = p * cos( 2.0 * theta * ONEDEG );
      u = p * sin( 2.0 * theta * ONEDEG ); */

      if( psig < 0.0 )
	psig = sqrt( pow( (q/p*qsig), 2.0) + pow( (u/p*usig), 2.0 ) );

      if( thetasig < 0.0 )
	thetasig = sqrt( 0.25 * ( ( pow( (u/(p*p)*qsig), 2.0 ) )\
		   + pow( (q/(p*p)*usig), 2.0 ) ) ) / ONEDEG;
      /*     thetasig = atan( psig / p ) / ONEDEG; */
      /*     thetasig = 90.0 / 3.1416 * psig / p; */

      while( theta > thetamax )
	theta -= 180.0;
      while( theta < thetamin )
	theta += 180.0;

      if( i < 1.0e-300 ){
	isig = 0.0;
	qsig = 0.0;
	usig = 0.0;
	psig = 0.0;
	thetasig = 0.0;
      }
      fprintf( fp, "%7.2f %10e %+7.3f %6.3f %+7.3f %6.3f %7.3f %6.3f %6.2f %6.2f %3.0f %7.2f %s %7.2f\n",\
	      lam_eff[pix], i, q, qsig, u, usig, p, psig, theta, thetasig,\
	      n_qu[pix], lam_c[pix], lam_w[pix], dlam_eff[pix] );
    }
    else{
      if( i < 1.0e-5 )
	isig = 0.0;
      lambda = crval + ( pix + 1.0 - crpix ) * cdelt;
      fprintf(fp, "%7.2f %10e %10e  %10e %10e\n", lambda, i, isig,\
	      ( i*yphot[pix]/exptime ), ( isig*yphot[pix]/exptime ) );
    }
  }

  fclose(fp);
    
  return( 0 );
}
